"""Implementation of FBA Problem class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np
import pandas as pd

from xbanalysis.problems.lp_problem import LpProblem


class FbaProblem:

    def __init__(self, xba_model):
        """Initialize FbaProblem from a xba_model.

        raise an exceptions, if fba parameters are not configured in the xba_model

        FBA mass balanced constraints: S * v = 0
        - rows of stoichiometric matrix S:
          - all metabolites (incl. boundary/constant metabolites),
          - i.e. excluding enzymes
        - columns of stoichiometric matrix S:
          - all reactions not containing enzymes as reactants or products
          - i.e. exlcluding protein synthesis, protein degradation, enzyme transitions
          - reversible reactions are not getting split into forward and backward reaction (for now)

        FBA flux bounds:
        - lower and upper bounds for reactions extracted from xba_model
        - getter and setter methods

        FBA objective: c.T * v:
        - extracted from xba_model (fbcObjective), including direction
        - getter and setter methods

        :param xba_model: xba model with all required FBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model

        self.metab_sids = [s.id for s in xba_model.species.values() if s.sboterm == 'SBO:0000247']
        metabs = set(self.metab_sids)
        # identify reactions that do not use enzymes in reactants and products
        self.mrids = [rid for rid, rxn in xba_model.reactions.items()
                      if len((set(rxn.products) | set(rxn.reactants)) - metabs) == 0]
        self.mrid2idx = {mrid: idx for idx, mrid in enumerate(self.mrids)}

        self.s_mat = xba_model.get_stoic_matrix(self.metab_sids, self.mrids)

        self.flux_bounds = np.array([[xba_model.reactions[rid].fbc_lb for rid in self.mrids],
                                     [xba_model.reactions[rid].fbc_ub for rid in self.mrids]])

        self.obj_coefs = np.zeros(len(self.mrids))
        for oid, fbc_obj in xba_model.fbc_objectives.items():
            if fbc_obj.active is True:
                self.obj_id = fbc_obj.id
                self.obj_dir = fbc_obj.direction
                self.set_obj_coefs(fbc_obj.coefficiants)
                break

    def create_fba_lp(self):
        """create Linear Programming Problem for FBA with split reactions

        Splitting reactions in forward/backward directions reduces high values during FBA
        Splitting required for pFBA (minimize absolute flux values)

        :return: LpProblem configured for FBA
        :rtype: LpProblem
        """
        lp = LpProblem()
        # split forward and backward reactions
        fwd_bwd_s_mat = np.hstack((self.s_mat, -self.s_mat))
        fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs))
        fwd_bwd_bounds = np.hstack((np.fmax(self.flux_bounds, 0),
                                    np.vstack((np.fmax(-self.flux_bounds[1], 0),
                                               np.fmax(-self.flux_bounds[0], 0)))))
        lp.configure(fwd_bwd_obj_coefs, fwd_bwd_bounds,
                     fwd_bwd_s_mat, np.zeros((2, fwd_bwd_s_mat.shape[0])), self.obj_dir)
        return lp

    def fba_optimize(self):
        """FBA optimization of current problem

        :return: results information from the optimization
        :rtype: dict
        """
        lp = self.create_fba_lp()
        res = lp.solve()
        del lp

        # combine results for split fluxes
        n_cols = len(self.mrids)
        res['x'] = res['x'][:n_cols] - res['x'][n_cols:]
        res['reduced_costs'] = res['reduced_costs'][:n_cols] - res['reduced_costs'][n_cols:]
        return res

    def pfba_optimize(self, fract_of_optimum=1.0):
        """ pfba analysis

        Reactions need to be split in forward/backard (to minimize absolute flux value)

        :param fract_of_optimum: value between 0.0 and 1.0.
        :type fract_of_optimum: float (default 1.0)
        :return:
        """
        lp = self.create_fba_lp()
        res = lp.solve(short_result=True)
        if res['simplex_status'] == 0:
            # fix FBA objective as constraint
            fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs))
            if self.obj_dir == 'maximize':
                row_bds = np.array([res['fun'] * fract_of_optimum, np.nan])
            else:
                row_bds = np.array([np.nan, res['fun'] / fract_of_optimum])
            lp.add_constraint(fwd_bwd_obj_coefs, row_bds)
            # new LP objective is to minimize sum of fluxes
            n_cols = len(self.mrids)
            lp.set_obj_coefs(np.ones(2 * n_cols))
            lp.set_obj_dir('minimize')
            res = lp.solve()
            # combine results for split fluxes
            res['x'] = res['x'][:n_cols] - res['x'][n_cols:]
            res['reduced_costs'] = res['reduced_costs'][:n_cols] - res['reduced_costs'][n_cols:]
        del lp
        return res

    def fva_optimize(self, rids=None, fract_of_optimum=1.0):
        """FVA - flux variability analysis for all or selected reactions.

        :param rids: selected reaction ids to run FVA (alternatively, all)
        :type rids: list of strings (default: None - all reactions)
        :param fract_of_optimum: value between 0.0 and 1.0.
        :type fract_of_optimum: float (default 1.0)
        :return: Flux ranges and objective values
        :rtype: pandas DataFrame
        """
        lp = self.create_fba_lp()
        res = lp.solve(short_result=True)

        if rids is None:
            rids = self.mrids
        flux_ranges = np.zeros((len(rids), 4))
        if res['simplex_status'] == 0:
            # fix FBA objective as constraint
            fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs))
            if self.obj_dir == 'maximize':
                row_bds = np.array([res['fun'] * fract_of_optimum, np.nan])
            else:
                row_bds = np.array([np.nan, res['fun'] / fract_of_optimum])
            fba_constr_row = lp.add_constraint(fwd_bwd_obj_coefs, row_bds)

            n_cols = len(self.mrids)
            for idx, rid in enumerate(rids):
                # select reaction maximize/minimize flux
                obj_coefs = np.zeros(2 * n_cols)
                obj_coefs[self.mrid2idx[rid]] = 1
                obj_coefs[self.mrid2idx[rid] + n_cols] = -1
                lp.set_obj_coefs(obj_coefs)

                lp.set_obj_dir('minimize')
                res = lp.solve(short_result=True)
                if res['simplex_status'] == 0:
                    flux_ranges[idx, 0] = res['fun']
                    flux_ranges[idx, 2] = lp.get_row_value(fba_constr_row)
                else:
                    print(f"FVA min failed for {rid}, {res['simplex_message']}")
                    flux_ranges[idx, 0] = np.nan

                lp.set_obj_dir('maximize')
                res = lp.solve(short_result=True)
                if res['simplex_status'] == 0:
                    flux_ranges[idx, 1] = res['fun']
                    flux_ranges[idx, 3] = lp.get_row_value(fba_constr_row)
                else:
                    print(f"FVA max failed for {rid}, {res['simplex_message']}")
                    flux_ranges[idx, 1] = np.nan
        del lp
        return pd.DataFrame(flux_ranges, index=rids, columns=['min', 'max', 'obj_min', 'obj_max'])

    @property
    def obj_dir(self):
        return self.__obj_dir

    @obj_dir.setter
    def obj_dir(self, dir_string):
        if 'min' in dir_string:
            self.__obj_dir = 'minimize'
        else:
            self.__obj_dir = 'maximize'

    def get_flux_bounds(self):
        """retrieve flux bounds for all reactions in a dict.

        :returns: flux bounds configured in the FBA problem
        :rtype: dict (key: reaction id, value: list with two float values for lower/upper bound)
        """
        return {mrid: [bds[0], bds[1]] for mrid, bds in zip(self.mrids, self.flux_bounds.T)}

    def set_flux_bounds(self, flux_bounds):
        """selectivly set lower and upper flux bounds for given reactions.

        :param flux_bounds: new lower and upper flux bounds selected reactions
        :type flux_bounds: dict (key: reaction id, value: list with two float values for lower/upper bound)
        """
        for rid, (lb, ub) in flux_bounds.items():
            self.flux_bounds[0, self.mrid2idx[rid]] = lb
            self.flux_bounds[1, self.mrid2idx[rid]] = ub

    def get_obj_coefs(self):
        """retrieve the coefficients of the FBA objective function.

        :returns: objective coefficients
        :rtype: dict (key: reaction id, value: float)
        """
        return {rid: self.obj_coefs[idx] for rid, idx in self.mrid2idx.items()
                if self.obj_coefs[idx] != 0.0}

    def set_obj_coefs(self, coefficiants):
        """set the coefficients of the FBA objective function.

        :param coefficiants: objective coefficients of selected reactions
        :type coefficiants: dict (key: reaction id, value: float)
        """
        self.obj_coefs = np.zeros(len(self.mrids))
        for rid, coef in coefficiants.items():
            self.obj_coefs[self.mrid2idx[rid]] = coef
