"""Implementation of LpProblem (linear programming problem) class.

built upon swiglpk

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np
import scipy
import scipy.sparse
import swiglpk as glpk


# status reported
simplex_status = {
    0: 'LP problem instance successfully solved',
    1: 'Unable to start search, initial basis specified in the '
       'problem object is invalid. Number basic variables is not same as the number rows.',
    2: 'Unable to start search, basis matrix corresponding to initial basis is singular.',
    3: 'Unable to start search, basis matrix corresponding to initial basis is ill-conditioned.',
    4: 'Unable to start search, some double-bounded variables have incorrect bounds.',
    5: 'Search prematurely terminated: solver failure.',
    6: 'Search prematurely terminated: objective function being '
       'maximized reached its lower limit and continues decreasing.',
    7: 'Search prematurely terminated: objective function being '
       'minimized reached its upper limit and continues increasing.',
    8: 'Search prematurely terminated: simplex iteration limit exceeded.',
    9: 'Search prematurely terminated: time limit exceeded.',
    10: 'LP problem instance has no primal feasible solution.',
    11: 'LP problem instance has no dual feasible solution.'}

generic_status = {1: 'solution is undefined',
                  2: 'solution is feasible',
                  3: 'solution is infeasible',
                  4: 'problem has no feasible solution', 
                  5: 'solution is optimal',
                  6: 'problem has unbounded solution'}

dual_status = {1: 'dual solution is undefined',
               2: 'dual solution is feasible',
               3: 'dual solution is infeasible',
               4: 'no dual feasible solution exists'}

prim_status = {1: 'primal solution is undefined',
               2: 'primal solution is feasible',
               3: 'primal solution is infeasible',
               4: 'no primal feasible solution exists'}

var_status = {1: 'basic variable',
              2: 'non-basic variable on its lower bound',
              3: 'non-basic variable on its upper bound',
              4: 'non-basic free (unbounded) variable',
              5: 'non-basic fixed variable'}

scaling_options = {'GM': glpk.GLP_SF_GM,
                   'EQ': glpk.GLP_SF_EQ,
                   '2N': glpk.GLP_SF_2N,
                   'SKIP': glpk.GLP_SF_SKIP,
                   'AUTO': glpk.GLP_SF_AUTO}


class LpProblem:

    def __init__(self):
        """create a glpk problem.

        Problem needs to be properly deleted usin del(LpProblem)
        """
        self.lp = glpk.glp_create_prob()
        self.ncols = 0
        self.nrows = 0
        self.simplex_result = None
        self.res = {}

    @staticmethod
    def _get_variable_type(bounds):
        """Determine variable type for upper/lower bounds.

        :param bounds: upper and lower bounds
        :type bounds: 1D ndarray of length 2
        :return var_type: glpk variable type to configure
        :rtype: integer constant
        :return lb: value of lower bound
        :rtype float
        :return ub: value of upper bound
        :rtype float
        """
        lb, ub = bounds
        if np.isfinite(lb) and np.isfinite(ub):
            if lb < ub:
                var_type = glpk.GLP_DB
            else:
                ub = 0.0
                var_type = glpk.GLP_FX
        elif not np.isfinite(lb):
            lb = 0.0
            var_type = glpk.GLP_UP
        elif not np.isfinite(ub):
            ub = 0.0
            var_type = glpk.GLP_LO
        else:
            lb = 0.0
            ub = 0.0
            var_type = glpk.GLP_FR
        return var_type, lb, ub

    def configure(self, obj_coefs, col_bnds, constr_coefs, row_bnds, direction='maximize'):
        """Configure the LP Problem.

        Note: glpk related indices start at 1 (not zero)
        :param obj_coefs: coefficients of objective function
        :type obj_coefs: 1D ndarray of shape (1, ncols)
        :param col_bnds: lower and upper bounds of structural (col) variables
        :type col_bnds: 2D ndarray of shape(2, ncols) (1st row: lower bounds, 2nd row: upper bounds)
        :param constr_coefs: matrix with constraint coefficients
        :type constr_coefs: 2D nparray with shape(nrows, ncols)
        :param row_bnds: lower and upper bounds of auxhiliary (row) variables
        :type row_bnds: 2D ndarray of shape(2, nrows) (1st row: lower bounds, 2nd row: upper bounds)
        :param direction: direction for the optimization ('maximize' or 'minimize')
        :type direction: string
        """
        self.nrows, self.ncols = constr_coefs.shape

        assert len(obj_coefs) == self.ncols
        assert col_bnds.shape == (2, self.ncols)
        assert row_bnds.shape == (2, self.nrows)

        # configure objective function
        glpk.glp_add_cols(self.lp, self.ncols)
        self.set_obj_coefs(obj_coefs)
        self.set_obj_dir(direction)

        # configure bounds on structural (columns) variables
        for i, lb_ub in enumerate(col_bnds.T):
            var_type, lb, ub = self._get_variable_type(lb_ub)
            glpk.glp_set_col_bnds(self.lp, 1 + i, var_type, lb, ub)

        # configure constraints
        glpk.glp_add_rows(self.lp, self.nrows)
        for i, lb_ub in enumerate(row_bnds.T):
            var_type, lb, ub = self._get_variable_type(lb_ub)
            glpk.glp_set_row_bnds(self.lp, 1 + i, var_type, lb, ub)
        constr_coefs_sparse = scipy.sparse.coo_matrix(constr_coefs)
        n_coefs = len(constr_coefs_sparse.row)
        ia = glpk.intArray(1 + n_coefs)
        ja = glpk.intArray(1 + n_coefs)
        ar = glpk.doubleArray(1 + n_coefs)
        for i in range(n_coefs):
            ia[1 + i] = int(1 + constr_coefs_sparse.row[i])
            ja[1 + i] = int(1 + constr_coefs_sparse.col[i])
            ar[1 + i] = constr_coefs_sparse.data[i]
        glpk.glp_load_matrix(self.lp, n_coefs, ia, ja, ar)

    def replace_constraint(self, row, constr_coefs, row_bnds):
        # set/replace one bounded constraint
        #
        # Parameters:
        # row:    row id to be replaced
        # coeffs: 1D np.array of floats of size number design variables
        #         coefficients != 0.0 will be configured
        # bounds: 1D numpy.ndarray with float of size two
        #         lists and tuples will be converted to numpy.ndarray
        #         lower_bound, upper_bound, np.nan and np.inf supported

        assert row <= glpk.glp_get_num_rows(self.lp)
        glpk.glp_set_row_bnds(self.lp, row, *self._get_variable_type(row_bnds))

        n_coef = len(np.nonzero(constr_coefs)[0])
        ind = glpk.intArray(1 + n_coef)
        val = glpk.doubleArray(1 + n_coef)
        idx = 1
        for i in np.nonzero(constr_coefs)[0]:
            ind[idx] = int(1 + i)
            val[idx] = constr_coefs[i]
            idx += 1
        glpk.glp_set_mat_row(self.lp, row, n_coef, ind, val)

    def add_constraint(self, constr_coefs, row_bnds):
        """add a single constraint to the lp problem

        :param constr_coefs: constraint coefficients for the new row
        :type constr_coefs: 1d ndarray of length ncols
        :param row_bnds: upper and lower bounds of the auxiliary (row) variable
        :type row_bnds: 1D ndarray of length 2
        :return:
        """
        row = glpk.glp_add_rows(self.lp, 1)
        self.replace_constraint(row, constr_coefs, row_bnds)
        return row

    def del_constraint(self, row):
        # delete one constraint
        #
        # Parameters:
        # row:    row id to be replaced
        assert row <= glpk.glp_get_num_rows(self.lp)
        num = glpk.intArray(1 + 1)
        num[1] = row
        glpk.glp_del_rows(self.lp, 1, num)

    def set_single_bound(self, react_idx, bounds):
        assert react_idx <= self.ncols
        if type(bounds) is not np.ndarray:
            bounds = np.array(bounds)
        glpk.glp_set_col_bnds(self.lp, 1 + react_idx, *self._get_variable_type(bounds))

    def set_obj_dir(self, direction):
        """Configure the optimization direction

        :param direction: direction for the optimization ('maximize' or 'minimize')
        :type direction: string
        """
        obj_dir = glpk.GLP_MAX if 'max' in direction else glpk.GLP_MIN
        glpk.glp_set_obj_dir(self.lp, obj_dir)

    def reset_cost(self, cost_coef=None):
        # reset coefficients of objective function (cols) to 0.0
        #
        # Parameters:
        #   cost_coef: None or 1d numpy.ndarray with floats
        #       None: reset all cost coefficiencs
        #       ndarray: reset only coest coefficiencs != 0
        if cost_coef is None:
            for i in range(glpk.glp_get_num_cols(self.lp)):
                glpk.glp_set_obj_coef(self.lp, int(1 + i), 0.0)
        else:
            for i in np.nonzero(cost_coef)[0]:
                glpk.glp_set_obj_coef(self.lp, int(1 + i), 0.0)

    def set_obj_coefs(self, obj_coefs):
        """set coefficients for objective function

        :param obj_coefs: coefficients of objective function
        :type obj_coefs: 1D ndarray of shape (1, ncols)
        """
        for i in range(self.ncols):
            glpk.glp_set_obj_coef(self.lp, int(1 + i), obj_coefs[i])

    def solve(self, short_result=False, scaling=None):
        """Solve the Linear Programming Problem.

        :param short_result: short result (only objective value and status) or long result
        :type short_result: bool (default:False)
        :param scaling: problem scaling options ('GM', 'EQ', '2N', 'SKIP', 'AUTO'
        :type scaling: string (default, None)
        :return: result information from the optimization stored in a dictionary
        :rtype: dict
        """
        if scaling is not None:
            opt = scaling_options.get(scaling, glpk.GLP_SF_AUTO)
            glpk.glp_scale_prob(self.lp, opt)
        # sjj = [glpk.glp_get_sjj(self.lp, 1+i) for i in range(glpk.glp_get_num_cols(self.lp))]
        # rii = [glpk.glp_get_rii(self.lp, 1+i) for i in range(glpk.glp_get_num_rows(self.lp))]
        self.simplex_result = glpk.glp_simplex(self.lp, None)
        return self.results(short_result)

    def get_row_value(self, row):
        """Retrieve right hand side of a constraint (auxiliary variable)

        :param row: index into constraints
        :type row: int
        :return: value of auxiliary varibale (value of rhs)
        :rtype: float
        """
        assert row <= glpk.glp_get_num_rows(self.lp)
        return glpk.glp_get_row_prim(self.lp, row)

    def results(self, short_result=False):
        """Collect results for the optimization

        Alternatively, reduced results, e.g. when doing pFBA

        :param short_result: short result (only objective value and status) or long result
        :type short_result: bool (default:False)
        :return: result from the FBA optimization
        :rtype: dict
        """
        self.res = {
            'fun': glpk.glp_get_obj_val(self.lp),
            'simplex_status': self.simplex_result,
            'simplex_message': simplex_status[self.simplex_result],
        }
        if short_result is False:
            self.res['x'] = np.array([glpk.glp_get_col_prim(self.lp, 1 + i) for i in range(self.ncols)])
            self.res['reduced_costs'] = np.array([glpk.glp_get_col_dual(self.lp, 1 + i) for i in range(self.ncols)])
            self.res['shadow_prices'] = np.array([glpk.glp_get_row_dual(self.lp, 1 + i) for i in range(self.nrows)])
            self.res['generic_status'] = generic_status[glpk.glp_get_status(self.lp)]
            self.res['primal_status'] = prim_status[glpk.glp_get_prim_stat(self.lp)]
            self.res['dual_status'] = dual_status[glpk.glp_get_dual_stat(self.lp)]
        return self.res

    def __del__(self):
        """Explicitly delete the LpProblem. (can we do without?)
        """
        if getattr(self, 'lp'):
            glpk.glp_delete_prob(self.lp)
